/* -*- mode: c; c-basic-offset: 4; -*- */
/* vi:set cindent ts=8 sts=4 sw=4: */
/*
 * Functions and data relating to the options.
 *
 * The functions are sorted alphabetically by name.
 *
 * All the options are hidden in this module, and can only be accessed through
 * the functions found here.
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <errno.h>
#include <ncurses.h>
#include <sys/stat.h>
#include <limits.h>
#include <ctype.h>
#include <unistd.h>
#include <sys/types.h>
#include <fcntl.h>
#include <assert.h>

#include "main.h"

#include "options.h"
#include "lib.h"
#include "misc.h"
#include "curs_lib.h"

/* option types */
enum {
    STRING_OPTION,
    PATH_OPTION,
    BOOLEAN_OPTION,
    NUMBER_OPTION, /* Signed, but always positive, integer */
    /* Invisible option, used to display blank lines in the options menu */
    NON_OPTION
};

struct option_entry {
    /* User-visible name of the option as it appears in the rc files, the
       options menu and when interacting with the user. */
    char *name;
    int type;			/* the type of option */
    bool initialized;           /* has the option been initialized? */
    /* we don't bother using a union containing the next six members */
    char *data;			/* STRING_OPTION and PATH_OPTION only */
    char *data_default;		/* STRING_OPTION and PATH_OPTION only */
    bool bit;			/* BOOLEAN_OPTION only */
    bool bit_default;		/* BOOLEAN_OPTION only */
    int value;			/* NUMBER_OPTION only */
    int value_default;		/* NUMBER_OPTION only */
    /* short description of the option, which is displayed on the message line
       when the option is highlighted in the options menu and which is written
       as a comment in the rc files */
    char *help;
};

/* Used by options_make_menu_entry() to align the entries in the
   options menu */
#define LONGEST_OPTION_NAME_NOT_INCLUDING_BOOLEAN_OPTIONS 19
/* The order of this initialization MUST correspond to the options
   enumeration in options.h. Options of type BOOLEAN_OPTION MUST be the
   last members of the array. */
struct option_entry option[] = {
    {
	"startup-working-dir", PATH_OPTION, 0, NULL, NULL, 0, 0, 0, 0,
	"The program's working directory at startup.",
    },
    {
	"filename-extensions", STRING_OPTION, 0, NULL, NULL, 0, 0, 0, 0,
	"Syntax example: .mp3:mp3player:.midi|midiplayer:.wav|wavplayer"
    },
    {
	"startup-load-dir", PATH_OPTION, 0, NULL, NULL, 0, 0, 0, 0,
	"Directory to load at startup (overrides startup-playlist)."
    },
    {
	"startup-playlist", PATH_OPTION, 0, NULL, NULL, 0, 0, 0, 0,
	"Playlist to load at startup if startup-load-dir is not given."
    },
    {
	"delay-between-songs", NUMBER_OPTION, 0, NULL, NULL, 0, 0, 0, 3,
	"Delay (in seconds) between each song when playing toggled songs."
    },
    /* Empty option, used as a blank line in the options menu to separate the
       boolean options from the others. */
    {
	"Error: you're not supposed to see this!",
	NON_OPTION, 0, NULL, NULL, 0, 0, 0, 0,
	"Error: you're not supposed to see this!"
    },
    {
	"skip-colors", BOOLEAN_OPTION, 0, NULL, NULL, 0, 0, 0, 0,
	"Don't use colors even if the terminal supports it."
    },
    {
	"show-path", BOOLEAN_OPTION, 0, NULL, NULL, 0, 0, 0, 0,
	"Show paths instead of filenames."
    },
    {
	"ignore-case", BOOLEAN_OPTION, 0, NULL, NULL, 0, 1, 0, 0,
	"Ignore case when sorting and searching."
    },
    {
	"confirm-quit-cmd", BOOLEAN_OPTION, 0, NULL, NULL, 0, 0, 0, 0,
	"Confirm the quit command."
    },

    /* Terminate the options, number_of_options() will stop at this element,
       when it sees that the first member is NULL. */
    {
	NULL, BOOLEAN_OPTION, 0, NULL, NULL, 0, 0, 0, 0,
	NULL
    },
};

/*
 * Edit the current option.
 */
void edit_option(int option_index, int *redraw, bool *option_changed)
{
    char buf[STRING];

    if (is_a_non_existing_option(option_index))
	return;

    if (is_a_boolean_option(option_index)) {
	msg("Boolean options can only be toggled.");
	return;
    }

    if (is_a_number_option(option_index)) {
	int i;

	if (enter_string(buf, sizeof(buf), EMPTY, NO_BROWSE, "%s: ",
			 option_name(option_index)) == NULL)
	    return;

	i = 0;
	clear_accumulator();
	while (isdigit(buf[i])) {
	    add_to_accumulator(buf[i]);
	    i++;
	}
	if (buf[i] != '\0') {
	    msg("Invalid number.");
	    return;
	}
	/* If `buf' is \0, accumulator_value() returns 0 */
	set_number_option(option_index, accumulator_value());
	*option_changed = TRUE;
	TURN_ON_FLAG(*redraw, REDRAW_CURRENT);
	return;
    }

    if (is_a_string_option(option_index)) {
	if (string_option_value(option_index) == NULL) {
	    if (enter_string(buf, sizeof(buf), EMPTY, NO_BROWSE, "%s: ",
			     option_name(option_index)) == NULL)
		return; /* The user aborted */
	} else {
	    snprintf(buf, sizeof(buf), string_option_value(option_index));
	    if (enter_string(buf, sizeof(buf), HALF_FILLED, NO_BROWSE, "%s: ",
			     option_name(option_index)) == NULL)
		return; /* The user aborted */
	}
	if (option_index == OPT_FILENAME_EXTENSIONS && buf[0] != '\0' &&
	    !has_the_correct_syntax_of_the_filename_extensions_variable(buf)) {
	    msg("Syntax error. Use the following syntax: "
		".ext1|player1:.ext2|player2");
	    return;
	}
	/* If `buf' is \0, the value of the option will be cleared. */
	set_string_option(option_index, buf);
	*option_changed = TRUE;
	TURN_ON_FLAG(*redraw, REDRAW_CURRENT);
	return;
    }

    if (is_a_path_option(option_index)) {
	/* We don't call enter_path(), since we want to be able to hit
	   just a linefeed, as we can with enter_string(). */
	if (path_option_value(option_index) == NULL) {
	    if (enter_string(buf, sizeof(buf), EMPTY, NO_BROWSE,
			     "%s: ", option_name(option_index)) == NULL)
		return; /* The user aborted */
	} else {
	    snprintf(buf, sizeof(buf), path_option_value(option_index));
	    if (enter_string(buf, sizeof(buf), HALF_FILLED, NO_BROWSE, "%s: ",
			     option_name(option_index)) == NULL)
		return; /* The user aborted */
	}
	if (buf[0] != '\0' && buf[0] != '/' && buf[0] != '~') {
	    msg("Only paths starting with ~ or / are allowed.");
	    return;
	}
	/* If `buf' is \0, the value of the option will be cleared. */
	set_path_option(option_index, buf);
	*option_changed = TRUE;
	TURN_ON_FLAG(*redraw, REDRAW_CURRENT);
	return;
    }

    assert(0); /* Unknown option */
}

/*
 * Return the option index of option_variable_name, or -1 if not found.
 * Does not care about non-existing options.
 */
int get_option_index_by_name(char *option_variable_name)
{
    int i;

    for (i = 0; option[i].name; i++)
	if (strcasecmp(option_variable_name, option[i].name) == 0)
	    return i;

    return -1;
}

/*
 * Initialize unset options. Must only be called once.
 */
void initialize_unset_options(void)
{
    int i;
    int total_number_of_options;
  /* make sure we only get called once (even though that doesn't hurt) */
#ifndef NDEBUG
    bool first = TRUE;
    assert(first);
    first = FALSE;
#endif /* NDEBUG */

    /* Ignore non-existing options */
    total_number_of_options = number_of_options();
    for (i = 0; i < total_number_of_options; i++) {
	if (option[i].initialized == FALSE) {
	  set_option_to_default_value(i);
	  mark_option_as_initialized(i);
	}
    }
}

/*
 * Return TRUE if option_index is a boolean option, else return FALSE.
 */
bool is_a_boolean_option(enum options option_index)
{
    return (option[option_index].type == BOOLEAN_OPTION ? TRUE : FALSE);
}

/*
 * Return TRUE if option_index is a non-existing option, else return FALSE.
 * Non-existing options are there to be used as empty lines in
 * the options menu, and the user should not know anything about them.
 */
bool is_a_non_existing_option(enum options option_index)
{
    return (option[option_index].type == NON_OPTION ? TRUE : FALSE);
}

/*
 * Return TRUE if option_index is a number option, else return FALSE.
 */
bool is_a_number_option(enum options option_index)
{
    return (option[option_index].type == NUMBER_OPTION ? TRUE : FALSE);
}

/*
 * Return TRUE if option_index is a path option, else return FALSE.
 */
bool is_a_path_option(enum options option_index)
{
    return (option[option_index].type == PATH_OPTION ? TRUE : FALSE);
}

/*
 * Return TRUE if option_index is a string option, else return FALSE.
 */
bool is_a_string_option(enum options option_index)
{
    return (option[option_index].type == STRING_OPTION ? TRUE : FALSE);
}

/*
 * Return TRUE if option_index is TRUE, else return FALSE.
 */
bool is_boolean_option_true(enum options option_index)
{
    assert(is_a_boolean_option(option_index));
    return option[option_index].bit; /* TRUE or FALSE */
}

/*
 * Return TRUE if the path option option_index contains a value, else return FALSE.
 */
bool is_path_option_set(enum options option_index)
{
    assert(is_a_path_option(option_index));
    return (option[option_index].data ? TRUE : FALSE);
}

/*
 * Return TRUE if the string option option_index contains a value, else return FALSE.
 */
bool is_string_option_set(enum options option_index)
{
    assert(is_a_string_option(option_index));
    return (option[option_index].data ? TRUE : FALSE);
}

/*
 * Mark the option option_index as initialized.
 * Does not care about non-existing options.
 */
void mark_option_as_initialized(enum options option_index)
{
  option[option_index].initialized = TRUE;
}

/*
 * Return the number of options.
 */
int number_of_options(void)
{
    int i = 0;

    while (option[i].name != NULL)
	i++; return i;
}

/*
 * Return the value of the number of option_index.
 */
int number_option_value(enum options option_index)
{
  assert(is_a_number_option(option_index));
  return option[option_index].value;
}

/*
 * Return pointer to static buffer containing the description for option_index.
 * Doesn't not care about non-existing options.
 */
char *option_description_string(enum options option_index)
{
	return option[option_index].help;
}

/*
 * Return pointer to static buffer containing the user visible name for
 * option_index.
 * The user visible name is the name the user sees in the options menu,
 * the rc files and so on. It is named similar to the members of the options
 * enumeration, eg. the user visible name for OPT_SOME_OPTION would be
 * some-option. Does not care about non-existing options.
 */
char *option_name(enum options option_index)
{
  	return option[option_index].name;
}

/*
 * Generate index line into str for the option whose number is index.
 */
void options_make_menu_entry(char *buf, size_t bufsize, int index)
{
#define LEN LONGEST_OPTION_NAME_NOT_INCLUDING_BOOLEAN_OPTIONS

    if (is_a_string_option(index) || is_a_path_option(index))
	snprintf(buf, bufsize, "%-*.*s = %s", LEN, LEN,
		 option_name(index),
	 /* here we should really use path_option_value() and string_option_value() */
		 option[index].data ? option[index].data : "<not set>");
    else if (is_a_boolean_option(index))
	snprintf(buf, bufsize, "%s %s",
		 is_boolean_option_true(index) ? "[*]" : "[ ]", option_name(index));
    else if (is_a_number_option(index))
	snprintf(buf, bufsize, "%-*.*s = %d", LEN, LEN,
		 option_name(index),
		 number_option_value(index));
    else if (is_a_non_existing_option(index))
	snprintf(buf, bufsize, "");
    else
	assert(0);
}

/*
 * Return pointer to static buffer containing the value of the path option
 * option_index.
 */
char *path_option_value(enum options option_index)
{
  assert(is_a_path_option(option_index));
	return option[option_index].data;
}

/*
 * Reset all options to their default values.
 */
void reset_all_options(void)
{
    int i;
    int total = number_of_options();

    for (i = 0; i < total; i++)
	set_option_to_default_value(i);
}

/*
 * Write options to file. Tell user about it
 * if `file' can't be opened for writing. If file already exists,
 * overwrite it.
 */
void save_options(char *file)
{
  FILE *fp;
  int i;
  int total_number_of_options;

  fp = fopen(file, "w");
  if (fp == NULL) {
    perror_msg("Unable to save options to %s", file);
    return;
  }
  fprintf(fp, "# %s version %s configuration file -- customize as needed.\n",
	  PROGNAME, VERSION);
  fprintf(fp,
	  "#\n"
	  "# Note! everything can be changed in a much better way through "
	  "the options menu.\n"
	  "# All variables are case insensitive.\n"
	  "# Boolean variables can be set to yes/no/true/false/on/off/1/0.\n"
	  "# Path variables must be start with tilde (~) or slash (/).\n"
	  "# Paths specified must exist at start up.\n"
	  "# You can add comments using pound signs (#).\n"
	  "# If a pound sign is found on a line, the rest of the line is "
	  "ignored.\n"
	  "# Warning! If you save the options from %s, the comments will "
	  "be overwritten.\n"
	  "# Lines ending with backslash (\\) are continued on the next "
	  "line.\n\n",
	  PROGNAME);

  total_number_of_options = number_of_options();
  for (i = 0; i < total_number_of_options; i++) {
    if (!is_a_non_existing_option(i)) {
      fprintf(fp, "# %s\n", option_description_string(i));
      fprintf(fp, "# Variable type: %s\n",
	      is_a_path_option(i) ? "path" :
	      is_a_string_option(i) ? "string" :
	      is_a_boolean_option(i) ? "boolean" :
	      is_a_number_option(i) ? "digit" :
	      "An unexpected error has occured, contact the author!");
    }
    if (is_a_string_option(i) || is_a_path_option(i))
      fprintf(fp, "%s = %s\n\n", option_name(i), option[i].data ? option[i].data : "");
    else if (is_a_boolean_option(i))
      fprintf(fp, "%s = %s\n\n", option_name(i), is_boolean_option_true(i) ? "yes" : "no");
    else if (is_a_number_option(i))
      fprintf(fp, "%s = %d\n\n", option_name(i), number_option_value(i));
    else if (!is_a_non_existing_option(i))
      assert(0);
  }

  (void) fclose(fp);
}

/*
 * Set the boolean option option_index to new_value.
 */
void set_boolean_option(enum options option_index, bool new_value)
{
    assert(is_a_boolean_option(option_index));
    option[option_index].bit = new_value;
}

/*
 * Set the number option option_index to n.
 */
void set_number_option(enum options option_index, int n)
{
    assert(is_a_number_option(option_index));
    option[option_index].value = n;
}

/*
 * Set option_index to its default value.
 */
void set_option_to_default_value(enum options option_index)
{
  if (is_a_path_option(option_index))
    set_path_option(option_index, option[option_index].data_default);
  else if (is_a_string_option(option_index))
    set_string_option(option_index, option[option_index].data_default);
  else if (is_a_boolean_option(option_index))
    set_boolean_option(option_index, option[option_index].bit_default);
  else if (is_a_number_option(option_index))
    set_number_option(option_index, option[option_index].value_default);
  else if (!is_a_non_existing_option(option_index))
    assert(0);
}

/*
 * Set the path option option_index to s.
 * If s is \0 or NULL, set the option to NULL.
 */
void set_path_option(enum options option_index, const char *s)
{
  assert(is_a_path_option(option_index));
  safe_free(option[option_index].data);
  option[option_index].data = safe_strdup(s);
}

/*
 * Set the string option option_index to s.
 * If s is \0 or NULL, set the option to NULL.
 */
void set_string_option(enum options option_index, const char *s)
{
    assert(is_a_string_option(option_index));
    safe_free(option[option_index].data);
    option[option_index].data = safe_strdup(s);
}

/*
 * Return pointer to a static buffer containing the value of the string
 * option option_index.
 */
char *string_option_value(enum options option_index)
{
  assert(is_a_string_option(option_index));
  return option[option_index].data;
}

/*
 * Toggle the boolean option option_index. If it is TRUE, set it to FALSE,
 * if it is FALSE, set it to TRUE.
 */
void toggle_boolean_option(enum options option_index)
{
    assert(is_a_boolean_option(option_index));
    option[option_index].bit = !option[option_index].bit;
}
